==> mutt-0.48.1-h7.4.tgz: mutt-0.48.1-h7.4/dirlist.c <==
/* Copyright (C) 1996 Michael R. Elkins.
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"

#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#define MAX_ITEMS 255

#define M_FOLDERFMT "---Mutt: Folder Mode [Directory: %s]"

struct folder_file {
  mode_t mode;
  char *name;
  char *desc;
};

/*
 * Global variables for this package.
 */
static struct {
  struct folder_file **entries;
  short max;
  short desclen;
  short namelen;
} LOCAL = { 0, 0, 0, 0 };

/*
 * Frees up the memory allocated for the local-global variables.
 */
static void destroy_global (void)
{
  int c;

  for (c = 0; LOCAL.entries[c]; c++) {
    safe_free((void **)&LOCAL.entries[c]->name);
    safe_free((void **)&LOCAL.entries[c]->desc);
    safe_free((void **)&LOCAL.entries[c]);
  }
  safe_free((void **)&LOCAL.entries);
}

static void sort_entries (void)
{
  struct folder_file **newtop = (struct folder_file **)safe_malloc(sizeof(struct folder_file *) * (LOCAL.max + 1));
  int i, j, k, newtopmax = 1;

  newtop[0] = LOCAL.entries[0];
  newtop[LOCAL.max] = 0;
  for (i=1; i<LOCAL.max; i++) {
    for (j=0; j<newtopmax; j++) {
      if (strcmp(LOCAL.entries[i]->name, newtop[j]->name) < 0) {
        for (k=newtopmax; k>j; k--)
          newtop[k] = newtop[k-1];
        newtop[j] = LOCAL.entries[i];
        newtopmax++;
        break;
      }
    }
    if (j == newtopmax)
      newtop[newtopmax++] = LOCAL.entries[i];
  }
  safe_free((void **)&LOCAL.entries);
  LOCAL.entries = newtop;
}

static int examine_directory (const char *d)
{
  struct stat s;
  DIR *dp;
  struct dirent *de;
  int l;
  char buffer[LONG_STRING];

  if (stat(d, &s) == -1) {
    mutt_error("Could not stat that directory!");
    return(-1);
  }

  if (!S_ISDIR(s.st_mode)) {
    mutt_error("Warning: %s is not a directory", d);
    sleep(2);
    return(-1);
  }

#ifdef USE_SETGID
  if (access(d, R_OK) == -1) {
    mutt_error("You do not have permission to read that directory!");
    return(-1);
  }
#endif /* USE_SETGID */

  LOCAL.max = 0;
  LOCAL.namelen = 0;
  LOCAL.desclen = 0;
  LOCAL.entries = (struct folder_file **)safe_malloc(sizeof(struct folder_file *) * (MAX_ITEMS + 1));
  dp = opendir(d);
  while ((de = readdir (dp)) != NULL) {
    if (strcmp (de->d_name, ".") == 0) continue;    /* we don't need . */
    
    sprintf(buffer, "%s/%s", d, de->d_name);
    stat(buffer, &s);                              /* get file-info */
    
    if ((! S_ISREG(s.st_mode)) && (! S_ISDIR(s.st_mode)))
      continue;
    
    if (S_ISDIR(s.st_mode))
      sprintf(buffer, "Directory");
    else {
      sprintf(buffer, "(%7d bytes)", (int)s.st_size);
    }
    
    LOCAL.entries[LOCAL.max] = (struct folder_file *)safe_malloc(sizeof(struct folder_file));
    LOCAL.entries[LOCAL.max]->mode = s.st_mode;
    LOCAL.entries[LOCAL.max]->name = safe_strdup(de->d_name);
    LOCAL.entries[LOCAL.max]->desc = safe_strdup(buffer);
    
    /* desclen and namelen is strlen of the longest strings */
    if ((l = strlen(de->d_name)) > LOCAL.namelen) LOCAL.namelen = l;
    if ((l = strlen(buffer)) > LOCAL.desclen) LOCAL.desclen = l;
    
    if (++(LOCAL.max) == MAX_ITEMS) {
      mutt_error("Warning: directory listing was too large");
      dprint(1,(debugfile, "folder_select(): execeded MAX_ITEMS.\n"));
      break;
    }
  }
  closedir(dp);  
  LOCAL.entries[LOCAL.max] = 0;
  sort_entries();
  return 0;
}

/*
 * This is the callback routine to make the file entries displayed on
 * the screen.
 */
void folder_makeEntry (char *s, int num, int slen)
{
  char fmt[SHORT_STRING];

  /*
   * the version of snprintf distributed with Mutt doesn't seem to understand
   * the format "%-*s", so I have to build the format string first...
   */
  snprintf(fmt, sizeof(fmt), "%%-%ds   %%-%ds", LOCAL.namelen, LOCAL.desclen);
  snprintf(s, slen, fmt, LOCAL.entries[num]->name, LOCAL.entries[num]->desc);
}

void folder_select (const char *d, char *f)
{
  char prefix[STRING], buffer[STRING];
  MUTTMENU *menu;
  
  *f = 0;
  strfcpy(prefix, d, sizeof(prefix));
  if (examine_directory(prefix) == -1) return;

  snprintf(buffer, sizeof(buffer), M_FOLDERFMT, prefix);
  menu = mutt_newMenu(buffer);
  menu->menu = MENU_FOLDER;
  menu->makeEntry = folder_makeEntry;
  menu->max = LOCAL.max;

  FOREVER {
    switch (mutt_menuLoop(menu)) {
    case OP_GENERIC_SELECT_ENTRY:
      if (S_ISDIR(LOCAL.entries[menu->current]->mode)) {
        if (strcmp(LOCAL.entries[menu->current]->name,"..") == 0) {
          if (strcmp("..", prefix+strlen(prefix)-2) == 0)
            strcat(prefix, "/..");
          else {
            char *p = strrchr(prefix, '/');

            if (strcmp(prefix, "/") == 0);
            else if (p)
              *p = 0;
            else
              strcpy(prefix, "/..");
          }
        }
        else
          sprintf(prefix+strlen(prefix), "/%s", LOCAL.entries[menu->current]->name);
        destroy_global();
        if (examine_directory(prefix) == -1) return;
	menu->current = 0;
	menu->top = 0;
	menu->max = LOCAL.max;
	safe_free((void **)&menu->title);
	snprintf(buffer, sizeof(buffer), M_FOLDERFMT, prefix);
	menu->title = safe_strdup(buffer);
        menu->redraw = REDRAW_FULL;
	break;
      }
      sprintf(f, "%s/%s", prefix, LOCAL.entries[menu->current]->name);
      /*
       * Fall through to OP_GENERIC_EXIT
       */
      
    case OP_GENERIC_EXIT:
      destroy_global();
      mutt_menuDestroy(&menu);
      return;
    }
  }
  /* not reached */
}

==> mutt-0.56-h8.1.tgz: mutt-0.56-h8.1/dirlist.c <==
/*
 * Copyright (C) 1996 Michael R. Elkins <me@cs.hmc.edu>
 * 
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 * 
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 * 
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include "mutt.h"
#include "mutt_curses.h"
#include "mutt_menu.h"
#include "keymap.h"

#include <stdlib.h>
#include <sys/types.h>
#include <dirent.h>
#include <string.h>
#include <unistd.h>
#include <sys/stat.h>

#define MAX_ITEMS 255

#define M_FOLDERFMT "---Mutt: Folder Mode [Directory: %s]"

struct folder_file {
  mode_t mode;
  char *name;
  char *desc;
};

/*
 * Global variables for this package.
 */
static struct {
  struct folder_file **entries;
  short max;
  short desclen;
  short namelen;
} LOCAL = { 0, 0, 0, 0 };

/*
 * Frees up the memory allocated for the local-global variables.
 */
static void destroy_global (void)
{
  int c;

  for (c = 0; LOCAL.entries[c]; c++)
  {
    safe_free((void **)&LOCAL.entries[c]->name);
    safe_free((void **)&LOCAL.entries[c]->desc);
    safe_free((void **)&LOCAL.entries[c]);
  }
  safe_free((void **)&LOCAL.entries);
}

static void sort_entries (void)
{
  struct folder_file **newtop = (struct folder_file **)safe_malloc(sizeof(struct folder_file *) * (LOCAL.max + 1));
  int i, j, k, newtopmax = 1;

  newtop[0] = LOCAL.entries[0];
  newtop[LOCAL.max] = 0;
  for (i=1; i<LOCAL.max; i++) {
    for (j=0; j<newtopmax; j++) {
      if (strcmp(LOCAL.entries[i]->name, newtop[j]->name) < 0) {
        for (k=newtopmax; k>j; k--)
          newtop[k] = newtop[k-1];
        newtop[j] = LOCAL.entries[i];
        newtopmax++;
        break;
      }
    }
    if (j == newtopmax)
      newtop[newtopmax++] = LOCAL.entries[i];
  }
  safe_free((void **)&LOCAL.entries);
  LOCAL.entries = newtop;
}

static int examine_directory (const char *d)
{
  struct stat s;
  DIR *dp;
  struct dirent *de;
  int l;
  char buffer[LONG_STRING];

  if (stat (d, &s) == -1)
  {
    mutt_perror ("stat");
    return (-1);
  }

  if (!S_ISDIR (s.st_mode))
  {
    mutt_error ("Warning: %s is not a directory", buffer);
    sleep (2);
    return (-1);
  }

  LOCAL.max = 0;
  LOCAL.namelen = 0;
  LOCAL.desclen = 0;
  LOCAL.entries = (struct folder_file **)safe_malloc(sizeof(struct folder_file *) * (MAX_ITEMS + 1));
  if ((dp = opendir (d)) == NULL)
    return (-1);
  while ((de = readdir (dp)) != NULL)
  {
    if (strcmp (de->d_name, ".") == 0) continue;    /* we don't need . */
    
    sprintf(buffer, "%s/%s", d, de->d_name);
    if (stat(buffer, &s) == -1)                 /* get file-info */
      continue;
    
    if ((! S_ISREG(s.st_mode)) && (! S_ISDIR(s.st_mode))) continue;
    
    if (S_ISDIR(s.st_mode))
      sprintf(buffer, "Directory");
    else
      sprintf(buffer, "(%7d bytes)", (int)s.st_size);
    
    LOCAL.entries[LOCAL.max] = (struct folder_file *)safe_malloc(sizeof(struct folder_file));
    LOCAL.entries[LOCAL.max]->mode = s.st_mode;
    LOCAL.entries[LOCAL.max]->name = safe_strdup(de->d_name);
    LOCAL.entries[LOCAL.max]->desc = safe_strdup(buffer);
    
    /* desclen and namelen is strlen of the longest strings */
    if ((l = strlen(de->d_name)) > LOCAL.namelen) LOCAL.namelen = l;
    if ((l = strlen(buffer)) > LOCAL.desclen) LOCAL.desclen = l;
    
    if (++(LOCAL.max) == MAX_ITEMS) {
      mutt_error("Warning: directory listing was too large");
      dprint(1,(debugfile, "folder_select(): execeded MAX_ITEMS.\n"));
      break;
    }
  }
  closedir(dp);  
  LOCAL.entries[LOCAL.max] = 0;
  sort_entries();
  return 0;
}

/*
 * This is the callback routine to make the file entries displayed on
 * the screen.
 */
void folder_makeEntry (char *s, int num, size_t slen)
{
  char fmt[SHORT_STRING];

  /*
   * the version of snprintf distributed with Mutt doesn't seem to understand
   * the format "%-*s", so I have to build the format string first...
   */
  snprintf(fmt, sizeof(fmt), "%%-%ds   %%-%ds", LOCAL.namelen, LOCAL.desclen);
  snprintf(s, slen, fmt, LOCAL.entries[num]->name, LOCAL.entries[num]->desc);
}

void folder_select (const char *d, char *f)
{
  char prefix[_POSIX_PATH_MAX], buffer[STRING];
  MUTTMENU *menu;
  
  *f = 0;
  strfcpy (prefix, d, sizeof (prefix));
  mutt_expand_path (prefix, sizeof (prefix));
  if (examine_directory (prefix) == -1)
    return;

  snprintf (buffer, sizeof (buffer), M_FOLDERFMT, prefix);
  menu = mutt_newMenu (buffer);
  menu->menu = MENU_FOLDER;
  menu->makeEntry = folder_makeEntry;
  menu->max = LOCAL.max;

  FOREVER
  {
    switch (mutt_menuLoop(menu))
    {
    case OP_GENERIC_SELECT_ENTRY:
      if (S_ISDIR(LOCAL.entries[menu->current]->mode))
      {
        if (strcmp(LOCAL.entries[menu->current]->name,"..") == 0)
	{
          if (strcmp("..", prefix+strlen(prefix)-2) == 0)
            strcat(prefix, "/..");
          else {
            char *p = strrchr(prefix, '/');

            if (strcmp(prefix, "/") == 0);
            else if (p)
              *p = 0;
            else
              strcpy(prefix, "/..");
          }
        }
        else
          sprintf(prefix+strlen(prefix), "/%s", LOCAL.entries[menu->current]->name);
        destroy_global();
        if (examine_directory(prefix) == -1)
	  return;
	menu->current = 0;
	menu->top = 0;
	menu->max = LOCAL.max;
	safe_free((void **)&menu->title);
	snprintf(buffer, sizeof(buffer), M_FOLDERFMT, prefix);
	menu->title = safe_strdup(buffer);
        menu->redraw = REDRAW_FULL;
	break;
      }
      sprintf(f, "%s/%s", prefix, LOCAL.entries[menu->current]->name);
      /*
       * Fall through to OP_GENERIC_EXIT
       */
      
    case OP_GENERIC_EXIT:
      destroy_global ();
      mutt_menuDestroy (&menu);
      return;
    }
  }
  /* not reached */
}
